/*

SimpleServer.cpp

Illustrates the minimum code required to create and send marker data using
the NatNet server.

*/

#include <stdio.h>
#include <tchar.h>
#include <conio.h>
#include <winsock2.h>
#include "NatNetTypes.h"
#include "NatNetServer.h"

#pragma warning( disable : 4996 )

void BuildDescription(sDataDescriptions* pDescription);
void SendDescription(sDataDescriptions* pDescription);
void FreeDescription(sDataDescriptions* pDescription);

void BuildFrame(long FrameNumber, sDataDescriptions* pDataDescriptions, sFrameOfMocapData* pOutFrame);
void SendFrame(sFrameOfMocapData* pFrame);
void FreeFrame(sFrameOfMocapData* pFrame);
int __cdecl RequestHandler(sPacket* pPacketIn, sPacket* pPacketOut,void* pUserData);

NatNetServer theServer;							// The NatNet Server 
sDataDescriptions descriptions;					// Describes what is sent (Marker sets, rigid bodies, etc)

int _tmain(int argc, _TCHAR* argv[])
{
	// Handles requests from the Client
    theServer.SetMessageResponseCallback( RequestHandler );	

	// Initialize NatNet server
	unsigned long lAddresses[10];
	int nAddresses;
	nAddresses = theServer.GetLocalIPAddresses(lAddresses, 10);	// get all ip addresses on this machine
	if(nAddresses < 1)
	{
		printf("Unable to detect local network.  Exiting");
		return 1;
	}
	char szIPAddress[128];
	theServer.IPAddress_LongToString(lAddresses[0], szIPAddress);	// use first IP detected
	int retCode = theServer.Initialize(szIPAddress);
	if(retCode != 0)
	{
		printf("Error initializing server.  See log for details.  Exiting");
		return 1;
	}

	// Create a MarkerSet description
	BuildDescription(&descriptions);

	// OK! Ready to stream data.  Listen for request from clients (RequestHandler())
	printf("Commands:\ns\tnext frame\nq\tquit\n\n");
	bool bExit = false;
	while(int c =_getch())
	{
		switch(c)
		{
			case 's':							// send frame
				sFrameOfMocapData frame;
				BuildFrame(1, &descriptions, &frame);
				SendFrame(&frame);
				FreeFrame(&frame);
				break;
			case 'q':
				bExit = true;		
				break;
			default:
				break;
		}

		if(bExit)
		{
			theServer.Uninitialize();
			FreeDescription(&descriptions);
			break;
		}
	}

	return ErrorCode_OK;
}

// RequestHandler receives requests from the Client.  Server's should
// build and return "request responses" packets in this same thread.
//
//   pPacketIn  - data packet from Client
//   pPacketOut - empty packet, to be filled in and returned to the
//				  Client as the "request response".
//
int __cdecl RequestHandler(sPacket* pPacketIn, sPacket* pPacketOut, void* pUserData)
{
	int iHandled = 1;	// handled

    switch (pPacketIn->iMessage)
    {
    case NAT_REQUEST_MODELDEF:
		printf("[SimpleServer] Received request for model definition.\n");
		theServer.PacketizeDataDescriptions(&descriptions, pPacketOut);
        break;

	case NAT_REQUEST_FRAMEOFDATA:
		{
		// note: Client does not typically poll for data, but we accomodate it here anyway
		// note: need to return response on same thread as caller
		printf("[SimpleServer] Received request for frame of data.\n");
		sFrameOfMocapData frame;
		BuildFrame(1, &descriptions, &frame);
		theServer.PacketizeFrameOfMocapData(&frame, pPacketOut);
		FreeFrame(&frame);
		}
        break;

    default:
        pPacketOut->iMessage = NAT_UNRECOGNIZED_REQUEST;
        pPacketOut->nDataBytes = 0;
		iHandled = 0;
        break;
    }

	return iHandled; // 0 = not handled, 1 = handled;
}

// Build a DataSet descriptipn (MarkerSets, RigiBodies, Skeletons).
// Note : This example currently just builds a MarkerSet.
void BuildDescription(sDataDescriptions* pDescription)
{
	sMarkerSetDescription* pMarkerSetDescription = new sMarkerSetDescription();

	strcpy(pMarkerSetDescription->szName, "Katie");
	pMarkerSetDescription->nMarkers = 10;
	pMarkerSetDescription->szMarkerNames = new char*[pMarkerSetDescription->nMarkers];
	char szTemp[128];
    for (int i=0; i < pMarkerSetDescription->nMarkers; i++)
	{
		sprintf(szTemp, "Marker %d", i);
		pMarkerSetDescription->szMarkerNames[i] = new char[MAX_NAMELENGTH];
		strcpy(pMarkerSetDescription->szMarkerNames[i], szTemp);
	}

	pDescription->nDataDescriptions = 1;
	pDescription->arrDataDescriptions[0].type = Descriptor_MarkerSet;
	pDescription->arrDataDescriptions[0].Data.MarkerSetDescription = pMarkerSetDescription;

}

// Send DataSet description to Client
void SendDescription(sDataDescriptions* pDescription)
{
	sPacket packet; 
	theServer.PacketizeDataDescriptions(pDescription, &packet);
	theServer.SendPacket(&packet);
}

// Release any memory we allocated for our example
void FreeDescription(sDataDescriptions* pDescription)
{
	sMarkerSetDescription* pMarkerSetDescription = pDescription->arrDataDescriptions[0].Data.MarkerSetDescription;
	for(int i=0; i< pMarkerSetDescription->nMarkers; i++)
	{
		delete[] pMarkerSetDescription->szMarkerNames[i];
	}
	delete[] pMarkerSetDescription->szMarkerNames;
}

// Build frame of MocapData
void BuildFrame(long FrameNumber, sDataDescriptions* pModels, sFrameOfMocapData* pOutFrame)
{	
	if(!pModels)
	{
		printf("No models defined - nothing to send.\n");
		return;
	}

	// build frame of marker data
	ZeroMemory(pOutFrame, sizeof(sFrameOfMocapData));
	sMarkerSetDescription* pMS = pModels->arrDataDescriptions[0].Data.MarkerSetDescription;
	strcpy(pOutFrame->MocapData[0].szName, pMS->szName);
	pOutFrame->iFrame = 1;
	pOutFrame->nMarkerSets = pModels->nDataDescriptions;
    pOutFrame->nOtherMarkers = 0;
	pOutFrame->fLatency = 66.6f;
	pOutFrame->MocapData[0].nMarkers = pMS->nMarkers;
	pOutFrame->MocapData[0].Markers = new MarkerData[pOutFrame->MocapData[0].nMarkers];
	for(int iMarker=0; iMarker < pOutFrame->MocapData[0].nMarkers; iMarker++)
	{	    
		pOutFrame->MocapData[0].Markers[iMarker][0] = iMarker + 0.1f;		// x
		pOutFrame->MocapData[0].Markers[iMarker][1] = iMarker + 0.2f;		// y
		pOutFrame->MocapData[0].Markers[iMarker][2] = iMarker + 0.3f;		// z
	}
}

// Packetize and send a single frame of mocap data to the Client
void SendFrame(sFrameOfMocapData* pFrame)
{
	sPacket packet; 
	theServer.PacketizeFrameOfMocapData(pFrame, &packet);
	theServer.SendPacket(&packet);
}

// Release any memory we allcoated for this sample
void FreeFrame(sFrameOfMocapData* pFrame)
{
	delete[] pFrame->MocapData[0].Markers;
}
